/*
 * Wiz platform code
 * written by Gražvydas "notaz" Ignotas
 *
 * This code comes under the same license as
 * all other code in this program (BSD).
 */
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <sys/mman.h>
#include <unistd.h>
#include <sys/ioctl.h>
#include <linux/fb.h>

#include "gp2x.h"
#include "squidgehack.h"

static volatile unsigned int *gp2x_memregl;
static int memdev, gpiodev;

void *gp2x_screen;

#define FB_BUF_COUNT	2
#define FB_BUF_SIZE	(320*240)

static unsigned int fb_paddr[FB_BUF_COUNT];
static void *fb_vaddr[FB_BUF_COUNT];
static unsigned int fb_work_buf;
static int fbdev;


void gp2x_video_flip(void)
{
  gp2x_memregl[0x406C>>2] = fb_paddr[fb_work_buf];
  gp2x_memregl[0x4058>>2] |= 0x10;
  fb_work_buf++;
  if (fb_work_buf >= FB_BUF_COUNT)
    fb_work_buf = 0;
  gp2x_screen = fb_vaddr[fb_work_buf];
}

void gp2x_video_changemode(int bpp)
{
  int code = 0, bytes = 2;
  unsigned int r;
  switch (bpp)
  {
    case 8:
      code = 0x443a;
      bytes = 1;
      break;

    case 15:
    case 16:
      code = 0x4432;
      bytes = 2;
      break;

    default:
      printf("unhandled bpp request: %d\n", bpp);
      return;
  }

  gp2x_memregl[0x405c>>2] = bytes;
  gp2x_memregl[0x4060>>2] = bytes * 320;

  r = gp2x_memregl[0x4058>>2];
  r = (r & 0xffff) | (code << 16) | 0x10;
  gp2x_memregl[0x4058>>2] = r;
}

void gp2x_video_setpalette(int *pal, int len)
{
  /* pollux palette is 16bpp only.. */
  int i;
  for (i = 0; i < len; i++)
  {
    int c = pal[i];
    c = ((c >> 8) & 0xf800) | ((c >> 5) & 0x07c0) | ((c >> 3) & 0x001f);
    gp2x_memregl[0x4070>>2] = (i << 24) | c;
  }
}

unsigned long gp2x_joystick_read(void)
{
  unsigned long value = 0;
  read(gpiodev, &value, 4);
  if (value & 0x02)
   value |= 0x05;
  if (value & 0x08)
   value |= 0x14;
  if (value & 0x20)
   value |= 0x50;
  if (value & 0x80)
   value |= 0x41;
  return value;
}

void gp2x_init(void)
{
  struct fb_fix_screeninfo fbfix;
  int i, ret;

  memdev = open("/dev/mem", O_RDWR);
  gp2x_memregl = mmap(0, 0x10000, PROT_READ|PROT_WRITE,
    MAP_SHARED, memdev, 0xc0000000);
  gpiodev = open("/dev/GPIO", O_RDONLY);
  fbdev = open("/dev/fb0", O_RDWR);

  if (memdev < 0 || gp2x_memregl == MAP_FAILED || gpiodev < 0 || fbdev < 0)
  {
    printf("init failed! (%d %p %d %d)", memdev, gp2x_memregl, gpiodev, fbdev);
    exit(1);
  }

  ret = ioctl(fbdev, FBIOGET_FSCREENINFO, &fbfix);
  if (ret == -1)
  {
    perror("ioctl(fbdev) failed");
    exit(1);
  }

  printf("framebuffer: \"%s\" @ %08lx\n", fbfix.id, fbfix.smem_start);
  fb_paddr[0] = fbfix.smem_start;

  fb_vaddr[0] = mmap(0, FB_BUF_SIZE * FB_BUF_COUNT, PROT_READ|PROT_WRITE,
    MAP_SHARED, memdev, fb_paddr[0]);
  if (fb_vaddr[0] == MAP_FAILED)
  {
    perror("mmap(fb_vaddr) failed");
    exit(1);
  }
  memset(fb_vaddr[0], 0, FB_BUF_SIZE * FB_BUF_COUNT);

  printf("  %p -> %08x\n", fb_vaddr[0], fb_paddr[0]);
  for (i = 1; i < FB_BUF_COUNT; i++)
  {
    fb_paddr[i] = fb_paddr[i-1] + FB_BUF_SIZE;
    fb_vaddr[i] = (char *)fb_vaddr[i-1] + FB_BUF_SIZE;
    printf("  %p -> %08x\n", fb_vaddr[i], fb_paddr[i]);
  }
  fb_work_buf = 0;

  gp2x_video_flip();
  mmuhack();
}

void gp2x_deinit(void)
{
  mmuunhack();
  gp2x_video_changemode(16);
  gp2x_memregl[0x406C>>2] = fb_paddr[0];
  gp2x_memregl[0x4058>>2] |= 0x10;

  close(fbdev);
  close(gpiodev);
  munmap(fb_vaddr[0], FB_BUF_SIZE * FB_BUF_COUNT);
  munmap((void *)gp2x_memregl, 0x10000);
  close(memdev);

  exit(0);
}

